/**
  *  \file planetscentral/mailout/template.cc
  *  \brief Mailout: Template Processing
  */

#include <stdexcept>
#include <vector>
#include "planetscentral/mailout/template.h"
#include "planetscentral/net/url.h"
#include "io/textfile.h"
#include "planetscentral/net/encoding.h"
#include "planetscentral/dbclient/connection.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/fileclient.h"


static string_t
eatWord(string_t& str)
{
    string_t::size_type i = str.find_first_not_of(" \t");
    if (i >= str.size()) {
        return string_t();
    }
    string_t::size_type j = str.find_first_of(" \t", i);
    string_t result;
    if (j >= str.size()) {
        result.assign(str, i, str.size()-i);
        str.clear();
    } else {
        result.assign(str, i, j-i);
        str.erase(0, j);
    }
    return result;
}


/** Generate an attachment.
    \param forUser user to use for authentication
    \param result generate the attachment here
    \param u URL to fetch */
static void
generateAttachment(string_t forUser, MimeBuilder& result, const Url& u)
{
    if (u.getProtocol() == "c2file") {
        // Parameters
        int port = u.getPort();
        if (port == 0) {
            port = 7776;
        }
        string_t user = u.getUser();
        if (user.empty()) {
            user = forUser;
        }

        // Do it
        DbClient conn;
        conn.connectRetry(u.getHost(), itoa(port));
        FileClient file(conn);
        file.setUserPermissions(user);

        string_t content = file.getFile(u.getPath().substr(1));
        result.addBase64(content.data(), content.size());
    } else {
        // We only speak c2file protocol so far
        throw std::runtime_error(string_t("unsupported protocol '" + u.getProtocol() + "' in attachment URL").c_str());
    }
}

static string_t
getMimeType(string_t basename)
{
    string_t::size_type i = basename.rfind('.');
    if (i < basename.size()) {
        /* This is the same repertoire as in file.cgi as of 02/Apr/2012. */
        string_t ext = strLCase(basename.substr(i+1));
        if (ext == "ini" || ext == "src" || ext == "txt" || ext == "cfg" || ext == "log" || ext == "q" || ext == "frag") {
            return "text/plain; charset=ISO-8859-1";
        } else if (ext == "html" || ext == "htm") {
            return "text/html";
        } else if (ext == "png") {
            return "image/png";
        } else if (ext == "gif") {
            return "image/gif";
        } else if (ext == "jpg" || ext == "jpeg") {
            return "image/jpeg";
        } else if (ext == "bmp") {
            return "image/bmp";
        } else if (ext == "zip") {
            return "application/zip";
        } else {
            return "application/octet-stream";
        }
    } else {
        return "application/octet-stream";
    }
}


/******************************** Template *******************************/

/** Constructor. */
Template::Template()
{ }

/** Add variable for expansion. */
void
Template::addVariable(string_t name, string_t value)
{
    variables[name] = value;
}

/** Attach a file by URL. */
void
Template::addFile(string_t url)
{
    attachments.push_back(url);
}

/** Generate the mail body.
    \param in Read template from this file
    \param forUser User to use for authenticating attachments
    \param smtpAddress SMTP mail address for "To" header
    \return newly-allocated, completely populated MimeBuilder with the mail body */
std::auto_ptr<MimeBuilder>
Template::generate(TextFile& in, string_t forUser, string_t smtpAddress)
{
    bool haveAttachments = !attachments.empty();

    // With no attachments, create single-part MimeBuilder
    std::auto_ptr<MimeBuilder> result(new MimeBuilder(haveAttachments ? "multipart/mixed" : ""));

    // Generate header.
    string_t bodyContentType = "text/plain; charset=UTF-8";
    string_t bodyCTE = "quoted-printable";
    string_t line;
    while (in.getLine(line) && !line.empty()) {
        if (line[0] == '!') {
            processCommand(line);
        } else {
            string_t::size_type i = line.find(':');
            if (i == line.npos) {
                throw std::runtime_error(string_t("syntax error in template line '" + line + "'").c_str());
            }
            string_t headerName(line, 0, i);
            string_t headerValue(strTrim(line.substr(i+1)));
            if (strCaseCmp("Content-Type", headerName) == 0) {
                bodyContentType = headerValue;
            } else if (strCaseCmp("Content-Transfer-Encoding", headerName) == 0) {
                bodyCTE = headerValue;
            } else {
                result->addHeader(headerName, encodeMimeHeader(expand(headerValue), "UTF-8"));
            }
        }
    }
    result->addHeader("To", smtpAddress);
    if (!haveAttachments) {
        result->addHeader("Content-Type", bodyContentType);
        result->addHeader("Content-Transfer-Encoding", bodyCTE);
    }

    // Generate text body
    if (haveAttachments) {
        result->addBoundary();
        result->addHeader("Content-Type", bodyContentType);
        result->addHeader("Content-Disposition", "inline");
        result->addHeader("Content-Transfer-Encoding", bodyCTE);
    }
    while (in.getLine(line)) {
        processLine(*result, line);
    }

    // Generate attachments
    if (haveAttachments) {
        for (std::list<string_t>::iterator p = attachments.begin(); p != attachments.end(); ++p) {
            // Parse URL
            Url u;
            if (!u.parse(*p)) {
                throw std::runtime_error(string_t("invalid attachment URL '" + *p + "'").c_str());
            }

            // Do it
            result->addBoundary();
            result->addHeader("Content-Type", getMimeType(u.getBaseName()));
            result->addHeader("Content-Disposition", "attachment; filename=\"" + u.getBaseName() + "\"");
            result->addHeader("Content-Transfer-Encoding", "base64");
            generateAttachment(forUser, *result, u);
        }

        result->addBoundary();
    }

    result->finish();
    return result;
}

void
Template::processLine(MimeBuilder& result, string_t text)
{
    if (text.empty()) {
        // Blank line, just add it.
        result.addLineQP(string_t());
    } else if (text[0] == '!') {
        // Command
        processCommand(text);
    } else {
        // If a nonempty line expands to empty text, ignore it.
        string_t x(expand(text));
        if (!x.empty()) {
            result.addLineQP(x);
        }
    }
}

void
Template::processCommand(string_t text)
{
    text.erase(0, 1);
    string_t keyword = eatWord(text);
    if (keyword.empty() || keyword[0] == '-') {
        // comment
    } else if (keyword == "set") {
        string_t name = expand(eatWord(text));
        string_t value = expand(strTrim(text));
        if (!name.empty()) {
            variables[name] = value;
        }
    } else {
        // huh?
        throw std::runtime_error(string_t("unsupported keyword '" + keyword + "' in template").c_str());
    }
}

string_t
Template::expand(const string_t text)
{
    std::vector<string_t> stack;
    stack.push_back(string_t());

    string_t::size_type n = 0;
    string_t::size_type p;
    while ((p = text.find_first_of("$)", n)) != string_t::npos) {
        stack.back().append(text, n, p-n);
        if (text[p] == ')') {
            if (stack.size() > 1) {
                std::map<string_t, string_t>::iterator i = variables.find(stack.back());
                stack.pop_back();
                if (i != variables.end()) {
                    stack.back().append(i->second);
                }
            } else {
                stack.back().append(1, ')');
            }
            ++p;
        } else {
            ++p;
            if (p < text.size()) {
                if (text[p] == '(') {
                    stack.push_back(string_t());
                } else {
                    stack.back().append(text, p, 1);
                }
                ++p;
            }
        }
        n = p;
    }
    stack.back().append(text, n, text.size()-n);
    return stack.front();
}
