/**
  *  \file planetscentral/talk/message.cc
  */

#include "planetscentral/talk/message.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/talk.h"
#include "cpluslib/format.h"
#include "planetscentral/talk/topic.h"
#include "int/array.h"
#include "cpluslib/misc.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/forum.h"

Message::Message(DbClient& client, int32_t messageId)
    : message(DbSubtree(client, MESSAGE_ROOT).subtree(messageId)),
      messageId(messageId)
{ }

DbHash
Message::header()
{
    return message.hashKey("header");
}

DbHashIntField
Message::topicId()
{
    return header().intField("thread");
}

DbHashIntField
Message::parentMessageId()
{
    return header().intField("parent");
}

DbHashIntField
Message::postTime()
{
    return header().intField("time");
}

DbHashIntField
Message::editTime()
{
    return header().intField("edittime");
}

DbHashStringField
Message::author()
{
    return header().stringField("author");
}

DbHashStringField
Message::subject()
{
    return header().stringField("subject");
}

DbHashStringField
Message::rfcMessageId()
{
    return header().stringField("msgid");
}

DbHashStringField
Message::rfcHeaders()
{
    return header().stringField("rfcheader");
}

DbHashIntField
Message::sequenceNumber()
{
    return header().intField("seq");
}

DbHashIntField
Message::previousSequenceNumber()
{
    return header().intField("prevseq");
}

DbHashStringField
Message::previousRfcMessageId()
{
    return header().stringField("prevmsgid");
}

DbString
Message::text()
{
    return message.stringKey("text");
}

bool
Message::exists()
{
    return header().exists();
}

void
Message::remove()
{
    // Remove from sets, so it becomes invisible
    Topic t(topic());
    Forum f(t.forum());
    t.messages().remove(messageId);
    f.messages().remove(messageId);
    User(message.getClient(), author().get()).postedMessages().remove(messageId);

    // Remove from NNTP side
    removeRfcMessageId(message.getClient(), rfcMessageId().get());

    // If the topic is now empty, remove it completely
    if (t.messages().empty()) {
        t.removeEmpty();
    }

    // Remove post
    text().remove();
    header().remove();
}

Topic
Message::topic()
{
    return Topic(message.getClient(), topicId().get());
}

int32_t
Message::getId() const
{
    return messageId;
}

IntValue*
Message::describe()
{
    // FIXME: can we use HMGET?
    const char*const fields[] = { "thread", "parent", "time", "edittime", "author", "subject" };
    Ptr<IntArrayData> ad = new IntArrayData();
    for (std::size_t i = 0; i < countof(fields); ++i) {
        ad->content.pushNew(makeStringValue(fields[i]));
        ad->content.pushNew(header().getFieldRaw(fields[i]));
    }
    ad->content.pushNew(makeStringValue("msgid"));
    ad->content.pushNew(makeStringValue(getRfcMessageId()));
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

string_t
Message::getRfcMessageId()
{
    string_t msgid = rfcMessageId().get();
    if (msgid.empty()) {
        msgid = format("%d.%d%s", messageId, sequenceNumber().get(), msgid_suffix);
    }
    return msgid;
}

string_t
Message::getPreviousRfcMessageId()
{
    string_t msgid = previousRfcMessageId().get();
    if (msgid.empty()) {
        msgid = format("%d.%d%s", messageId, previousSequenceNumber().get(), msgid_suffix);
    }
    return msgid;
}

int32_t
Message::allocateMessage(DbClient& client)
{
    return ++DbSubtree(client, MESSAGE_ROOT).intKey("id");
}

int32_t
Message::getLastId(DbClient& client)
{
    return DbSubtree(client, MESSAGE_ROOT).intKey("id").get();
}

/** Remove the link from a RFC message Id to a posting. */
void
Message::removeRfcMessageId(DbClient& client, string_t id)
{
    if (!id.empty()) {
        DbSubtree(client, RFC_MSGID_ROOT).intKey(id).remove();
    }
}

/** Add a link from a RFC message Id to a posting. */
void
Message::addRfcMessageId(DbClient& client, string_t id, int32_t messageId)
{
    if (!id.empty() && messageId != 0) {
        DbSubtree(client, RFC_MSGID_ROOT).intKey(id).set(messageId);
    }
}
